def _is_in_bad_fork():
  return False


def manual_seed_all(seed):
  pass


def device_count():
  return 1


def get_rng_state():
  return []


def set_rng_state(new_state, device):
  pass


def is_available():
  return True


def current_device():
  return 0
